<?php
$INC = isset($INC) ? $INC : dirname(__FILE__);
require_once($INC . "/util.inc.php");
require_once($INC . "/log.inc.php");
require_once($INC . "/sql.inc.php");
require_once($INC . "/events.inc.php");

$rg_totp_error = '';

function rg_totp_set_error($str)
{
	global $rg_totp_error;
	$rg_totp_error = $str;
	rg_log('totp_set_error: ' . $str);
}

function rg_totp_error()
{
	global $rg_totp_error;
	return $rg_totp_error;
}

/*
 * Event functions
 */
$rg_totp_functions = array(
	//2006 => "rg_totp_link_by_name"
);
rg_event_register_functions($rg_totp_functions);


$rg_totp_base32_tab = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";

/*
 * Generates a secret (returned as google base32 string)
 */
function rg_totp_base32_generate($digits)
{
	global $rg_totp_base32_tab;

	$ret = '';
	for ($i = 0; $i < $digits; $i++)
		$ret .= $rg_totp_base32_tab[rand(0, 31)];

	return $ret;
}

/*
 * Transforms a base32 google string into a binary array
 */
function rg_totp_base32_decode($s)
{
	global $rg_totp_base32_tab;

	$s = strtoupper($s);
	$s_len = strlen($s);

	$ret = '';
	$buf = 0;
	$buf_bits = 0;
	$i = 0;
	while ($i < $s_len) {
		while (($i < $s_len) && ($buf_bits < 8)) {
			$n = strpos($rg_totp_base32_tab, $s[$i++]);
			$buf = ($buf << 5) | $n;
			$buf_bits += 5;
		}

		// less than 8 bits left, pad right with 0
		if ($buf_bits < 8) {
			$buf <<= (8 - $buf_bits);
			$buf_bits = 8;
		}

		// Now, we have more than 8 bits in buffer, extract
		// next out char.
		$buf_bits -= 8;
		$c = chr($buf >> $buf_bits);
		$ret .= $c;
		$buf &= ((1 << $buf_bits) - 1);
	}

	return $ret;
}

function rg_totp_compute($key, $tc, $digits)
{
	$key_bin = rg_totp_base32_decode($key);

	$tc_bin = hex2bin(sprintf("%016x", $tc));

	$h = hash_hmac('sha1', $tc_bin, $key_bin, TRUE /*bin output*/);

	$o = ord($h[strlen($h) - 1]) & 0x0F;
	$i = (ord($h[$o]) << 24) | (ord($h[$o + 1]) << 16)
		| (ord($h[$o + 2]) << 8) | ord($h[$o + 3]);
	$i &= 0x7FFFFFFF;
	return sprintf("%0" . $digits . "u", $i % pow(10, $digits));
}

/*
 * Verifies a tokens based on a specified 'tc'
 * Returns 'tc' if ok, FALSE on error
 */
function rg_totp_verify_tc($key, $tc, $token)
{
	$t = rg_totp_compute($key, $tc, 6);
	//rg_log('DEBUG: compute[tc=' . $tc . ']=' . $t);
	if (strcmp($token, $t) == 0)
		return $tc;

	return FALSE;
}

/*
 * Verifies if a login token is valid
 * We try 2 tcs before current ts and after current ts
 * Returns FALSE if token is not valid or 'tc' of the token if is valid.
 */
function rg_totp_verify($key, $ts, $token)
{
	rg_prof_start('totp_verify');
	rg_log_enter('totp_verify ts=' . $ts . ', token=' . $token);

	$tc = intval($ts / 30);
	$list = array($tc, $tc - 1, $tc - 2, $tc + 1, $tc + 2);
	foreach ($list as $tc) {
		$ret = rg_totp_verify_tc($key, $tc, $token);
		//rg_log('DEBUG: using tc ' . $tc . ', ret=' . $ret);
		if ($ret !== FALSE)
			break;
	}

	rg_log_exit();
	rg_prof_end('totp_verify');
	return $ret;
}

/*
 * Returns a PNG encoded QR code
 */
function rg_totp_png($secret)
{
	global $rg_ssh_host;

	$extra = gmdate('Y-m-d H:i');
	$issuer = $rg_ssh_host;
	$cmd = "qrencode -o - --level=H --type=PNG 'otpauth://totp/$extra?secret=$secret&issuer=$issuer'";
	$a = rg_exec($cmd);
	if ($a['ok'] != 1)
		return FALSE;

	return $a['data'];
}

/*
 * Returns a UTF-8 encoded QR code
 */
function rg_totp_text($secret)
{
	global $rg_ssh_host;

	$extra = gmdate('Y-m-d H:i');
	$issuer = $rg_ssh_host;
	$cmd = "qrencode -o - --level=M --margin=2 --type=UTF8 'otpauth://totp/$extra?secret=$secret&issuer=$issuer'";
	$a = rg_exec($cmd);
	if ($a['ok'] != 1)
		return FALSE;

	return $a['data'];
}

/*
 * Cosmetic fixes for a login token row
 */
function rg_totp_cosmetic(&$row)
{
	if (isset($row['itime']))
		$row['itime_nice'] = gmdate('Y-m-d H:i', $row['itime']);

	if (!isset($row['used']))
		$row['used'] = 0;

	if ($row['used'] > 0)
		$row['used_nice'] = gmdate('Y-m-d H:i', $row['used']);
	else
		$row['used_nice'] = "n/a";

	if (isset($row['conf'])) {
		if (strcmp($row['conf'], 't') == 0)
			$row['conf_nice'] = 'confirmed';
		else
			$row['conf_nice'] = 'pending';
	}
}

/*
 * Sets when a login token was last used
 */
function rg_totp_set_last_use($db, $uid, $id, $tc, $ts)
{
	rg_prof_start('totp_set_last_use');
	rg_log_enter('totp_set_last_use uid=' . $uid . ' id=' . $id
		. ' tc=' . $tc);

	$ret = FALSE;
	while (1) {
		$params = array('uid' => $uid,
			'id' => $id,
			'used' => $ts,
			'last_used_tc' => $tc,
			'conf' => 't');
		$sql = 'UPDATE login_tokens SET used = @@used@@'
			. ', last_used_tc = @@last_used_tc@@'
			. ', conf = @@conf@@'
			. ' WHERE uid = @@uid@@'
			. ' AND id = @@id@@';
                $res = rg_sql_query_params($db, $sql, $params);
                if ($res === FALSE) {
                        rg_user_set_error('cannot update last used (' . rg_sql_error());
			break;
                }
                rg_sql_free_result($res);

		$key = 'user' . '::' . $uid . '::' . 'login_tokens'
			. '::' . 'list' . '::' . $id;
		$a = array('used' => $ts, 'last_used_tc' => $tc, 'conf' => 't');
		rg_totp_cosmetic($a);
		rg_cache_merge($key, $a, RG_SOCKET_NO_WAIT);

		$ret = TRUE;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_set_last_use');
	return $ret;
}

/*
 * Returns a list of login tokens from database
 */
function rg_totp_list($db, $uid)
{
	rg_prof_start('totp_list');
	rg_log_enter('totp_list');

	while (1) {
		$key = 'user' . '::' . $uid . '::' . 'login_tokens';
		$ret = rg_cache_get($key);
		if ($ret !== FALSE) {
			$ret['ok'] = 1;
			break;
		}

		$ret = array();
		$ret['ok'] = 0;

		$params = array('uid' => $uid);
		$sql = 'SELECT * FROM login_tokens'
			. ' WHERE uid = @@uid@@'
			. ' ORDER BY itime';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot load login tokens');
			break;
		}

		$ret['list'] = array();
		while (($row = rg_sql_fetch_array($res))) {
			$id = $row['id'];

			rg_totp_cosmetic($row);

			$ret['list'][$id] = $row;
		}
		rg_sql_free_result($res);

		unset($ret['ok']); // we do not need it in cache
		rg_cache_set($key, $ret, RG_SOCKET_NO_WAIT);
		$ret['ok'] = 1;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_list');
	return $ret;
}

/*
 * Validates a token
 * Also, it marks the tokens as 'confirmed' if needed
 */
function rg_totp_verify_all($db, $uid, $token)
{
	rg_prof_start('totp_verify_all');
	rg_log_enter('totp_verify_all token=' . $token);

	$now = time();

	$ret = array();
	$ret['ok'] = 0;
	while (1) {
		$lt = rg_totp_list($db, $uid);
		if ($lt['ok'] != 1)
			break;

		$ret['ok'] = 1;
		$ret['enrolled'] = 0;
		$ret['id'] = 0;
		$ret['conf_error'] = 0;
		$ret['tc'] = 0;

		if (strlen($token) == 0) {
			rg_totp_set_error('token is empty');
			break;
		}
		$token = sprintf("%06u", $token);

		rg_totp_set_error('wrong token; sync the time');
		foreach ($lt['list'] as $t) {
			if (strcmp($t['conf'], 't') == 0)
				if ($ret['enrolled'] == 0)
					$ret['enrolled'] = 1;

			$tc = rg_totp_verify($t['secret'], $now, $token);
			if ($tc === FALSE)
				continue;

			if ($tc <= $t['last_used_tc']) {
				rg_totp_set_error('cannot reuse the login token');
				break;
			}

			$ret['id'] = $t['id'];
			$ret['tc'] = $tc;

			// Mark it as used and update 'conf' status
			$r = rg_totp_set_last_use($db, $uid, $t['id'], $tc, $now);
			if ($r !== TRUE) {
				$ret['ok'] = 0;
				break;
			}

			// we just confirmed an unconf entry, so we are enrolled
			if ($ret['enrolled'] == 0)
				$ret['enrolled'] = 1;
			break;
		}

		break;
	}

	rg_log_exit();
	rg_prof_end('totp_verify_all');
	return $ret;
}

/*
 * Add a new secret login token to database
 */
function rg_totp_enroll($db, $uid, $name, $secret, $ip, $conf)
{
	rg_prof_start('totp_enroll');
	rg_log_enter('totp_enroll name=' . $name);

	$ret = FALSE;
	while (1) {
		$params = array('uid' => $uid, 'name' => $name,
			'secret' => $secret,
			'itime' => time(), 'ip' => $ip, 'conf' => $conf,
			'last_used_tc' => 0);
		$sql = 'INSERT INTO login_tokens (uid, itime, name, secret, ip'
			. ', conf, last_used_tc)'
			. ' VALUES (@@uid@@, @@itime@@, @@name@@'
			. ', @@secret@@, @@ip@@, @@conf@@, @@last_used_tc@@)'
			. ' RETURNING id';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot insert login token; try again later');
			break;
		}
		$row = rg_sql_fetch_array($res);
                rg_sql_free_result($res);

		$params['id'] = $row['id'];
		rg_totp_cosmetic($params);
		$key = 'user' . '::' . $uid . '::' . 'login_tokens'
			. '::' . 'list' . '::' . $params['id'];
		rg_cache_set($key, $params, RG_SOCKET_NO_WAIT);

		$ret = TRUE;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_enroll');
	return $ret;
}

/*
 * Add an ip to login_tokens_ip table
 */
function rg_totp_add_ip($db, $uid, $token_id, $ip, $expire_ts)
{
	rg_prof_start('totp_add_ip');
	rg_log_enter('totp_add_ip ip=' . $ip . ' expire_ts=' . $expire_ts);

	$ret = FALSE;
	while (1) {
		$params = array('uid' => $uid,
			'token_id' => $token_id,
			'ip' => $ip,
			'itime' => time(),
			'expire' => $expire_ts);
		$sql = 'INSERT INTO login_tokens_ip'
			. ' (uid, ip, itime, expire, token_id)'
			. ' VALUES (@@uid@@, @@ip@@, @@itime@@, @@expire@@'
			. ', @@token_id@@)';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot insert login token ip; try again later');
			break;
		}
                rg_sql_free_result($res);

		$key = 'user' . '::' . $uid . '::' . 'login_tokens_ip';
		rg_cache_apush($key, $params, RG_SOCKET_NO_WAIT);

		$ret = TRUE;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_add_ip');
	return $ret;
}

/*
 * Del an ip from login_tokens_ip table
 */
function rg_totp_del_ip($db, $uid, $ip)
{
	rg_prof_start('totp_del_ip');
	rg_log_enter('totp_del_ip ip=' . $ip);

	$ret = FALSE;
	while (1) {
		$params = array('uid' => $uid, 'ip' => $ip);
		if (strcasecmp($ip, 'all') == 0) {
			$sql = 'DELETE FROM login_tokens_ip'
				. ' WHERE uid = @@uid@@';
		} else {
			$sql = 'DELETE FROM login_tokens_ip'
				. ' WHERE uid = @@uid@@'
				. ' AND ip = @@ip@@';
		}
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot delete ip; try again later');
			break;
		}
		$aff = rg_sql_affected_rows($res);
                rg_sql_free_result($res);

		if ($aff == 0) {
			rg_totp_set_error('ip not found');
			break;
		}

		$key = 'user' . '::' . $uid . '::' . 'login_tokens_ip';
		rg_cache_unset($key, RG_SOCKET_NO_WAIT);

		$ret = TRUE;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_del_ip');
	return $ret;
}

/*
 * Remove a list of login tokens
 */
function rg_totp_remove($db, $uid, $list)
{
	rg_prof_start('totp_remove');
	rg_log_enter('totp_remove uid=' . $uid
		. ' list=' . rg_array2string($list));

	$ret = FALSE;
	while (1) {
		if (empty($list)) {
			rg_totp_set_error('you did not select anything');
			break;
		}

		$my_list = array();
		foreach ($list as $id => $junk)
			$my_list[] = sprintf("%u", $id);

		$params = array('uid' => $uid);
		$sql_list = implode(', ', $my_list);

		$sql = 'DELETE FROM login_tokens_ip'
			. ' WHERE uid = @@uid@@'
			. ' AND token_id IN (' . $sql_list . ')';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot remove login tokens ips');
			break;
		}
		rg_sql_free_result($res);

		$sql = 'DELETE FROM login_tokens'
			. ' WHERE uid = @@uid@@'
			. ' AND id IN (' . $sql_list . ')';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot remove login token');
			break;
		}
		rg_sql_free_result($res);

		foreach ($my_list as $junk => $_id) {
			$key = 'user' . '::' . $uid . '::' . 'login_tokens'
				. '::' . 'list' . $_id;
			rg_cache_unset($key, RG_SOCKET_NO_WAIT);
		}

		$ret = TRUE;
		break;
	}

        rg_log_exit();
	rg_prof_end('totp_remove');
	return $ret;
}

/*
 * Returns a list of login tokens IPs from database
 */
function rg_totp_list_ip($db, $uid)
{
	rg_prof_start('totp_list_ip');
	rg_log_enter('totp_list_ip');

	while (1) {
		$key = 'user' . '::' . $uid . '::' . 'login_tokens_ip';
		$list = rg_cache_get($key);
		if ($list !== FALSE) {
			$ret['list'] = $list;
			$ret['ok'] = 1;
			break;
		}

		$ret = array();
		$ret['ok'] = 0;

		$params = array('uid' => $uid, 'now' => time());
		$sql = 'SELECT * FROM login_tokens_ip'
			. ' WHERE uid = @@uid@@'
			. ' AND expire >= @@now@@'
			. ' ORDER BY itime';
		$res = rg_sql_query_params($db, $sql, $params);
		if ($res === FALSE) {
			rg_totp_set_error('cannot load login tokens ip');
			break;
		}

		$ret['list'] = array();
		while (($row = rg_sql_fetch_array($res))) {
			unset($row['uid']);
			$ret['list'][] = $row;
		}
		rg_sql_free_result($res);

		rg_cache_set($key, $ret['list'], RG_SOCKET_NO_WAIT);
		$ret['ok'] = 1;
		break;
	}

	rg_log_exit();
	rg_prof_end('totp_list_ip');
	return $ret;
}

/*
 * Verifies that an IP is in the 'validated' list
 * Returns -1 on error, 0 if no IP were present and 1 if all is OK
 */
function rg_totp_verify_ip($db, $uid, $ip)
{
	rg_prof_start('totp_verify_ip');
	rg_log_enter('totp_verify_ip ip=' . $ip);

	$ret = array();
	$ret['ok'] = -1;
	$ret['enrolled'] = 0;
	while (1) {
		// we will return OK if user is not enrolled
		$r = rg_totp_list($db, $uid);
		if ($r['ok'] != 1)
			break;
		if (empty($r['list'])) {
			$ret['ok'] = 1;
			break;
		}

		// let's see if we enrolled
		foreach ($r['list'] as $t) {
			if (strcmp($t['conf'], 't') == 0) {
				$ret['enrolled'] = 1;
				break;
			}
		}

		$r = rg_totp_list_ip($db, $uid);
		if ($r['ok'] != 1)
			break;

		$ret['ok'] = 0;
		foreach ($r['list'] as $t) {
			if (strcasecmp($t['ip'], $ip) == 0) {
				$ret['list'] = $r['list'];
				$ret['ok'] = 1;
				break;
			}
		}

		if ($ret['ok'] == 0)
			rg_totp_set_error('you have no IP validated;'
				. ' run \'ssh ... totp\' for help');

		break;
	}

	rg_log_exit();
	rg_prof_end('totp_verify_ip');
	return $ret;
}

/*
 * High-level function for listing tokens
 */
function rg_totp_list_high_level($db, $rg)
{
	$ret = '';

	$del_errmsg = array();
	$rg['HTML:del_errmsg'] = '';
	$rg['HTML:del_status'] = '';

	$delete = rg_var_uint('delete');
	while ($delete == 1) {
		if (!rg_valid_referer()) {
			$del_errmsg[] = 'invalid referer; try again';
			break;
		}

		if (!rg_token_valid($db, $rg, 'login_tokens_list', FALSE)) {
			$del_errmsg[] = 'invalid token; try again.';
			break;
		}

		$list = rg_var_str("delete_list");
		$r = rg_totp_remove($db, $rg['login_ui']['uid'], $list);
		if ($r !== TRUE) {
			$del_errmsg[] = 'cannot delete: ' . rg_totp_error();
			break;
		}

		$rg['HTML:del_status'] = rg_template('user/settings/totp/delete_ok.html',
			$rg, TRUE /*xss*/);
		break;
	}

	$r = rg_totp_list($db, $rg['login_ui']['uid']);
	if ($r['ok'] !== 1) {
		$rg['totp_errmsg'] = rg_totp_error();
		$ret .= rg_template('user/settings/totp/list_err.html',
			$rg, TRUE /*xss*/);
	} else {
		$rg['rg_form_token'] = rg_token_get($db, $rg, 'login_tokens_list');
		$rg['HTML:del_errmsg'] = rg_template_errmsg($del_errmsg);
		$ret .= rg_template_table('user/settings/totp/list', $r['list'], $rg);
	}

	return $ret;
}

/*
 * Main function for TOTP login token
 */
function rg_totp_high_level($db, $rg)
{
	rg_prof_start('totp_high_level');
	rg_log_enter('totp_high_level');

	$now = time();

	$rg['totp'] = array();

	$ret = '';
	$errmsg = array();

	$rg['totp']['name'] = rg_var_str('totp::name');

	$enroll = rg_var_uint('enroll');
	while ($enroll == 1) {
		$name = rg_var_str('totp::name');
		$ver = rg_var_str('totp::ver');
		$secret = rg_var_str('totp::secret');

		if (strlen($name) == 0) {
			$errmsg[] = "invalid name";
			break;
		}

		if (strlen($ver) != 6) {
			$errmsg[] = "invalid number; you must enter a 6 digit number";
			break;
		}

		if (!rg_valid_referer()) {
			$errmsg[] = "invalid referer; try again";
			break;
		}

		if (!rg_token_valid($db, $rg, 'user_totp', FALSE)) {
			$errmsg[] = "invalid token; try again";
			break;
		}

		$r = rg_totp_verify($secret, $now, $ver);
		if ($r === FALSE) {
			$errmsg[] = rg_template('user/settings/totp/ver_error.html',
				$rg, TRUE /*xss*/);
			break;
		}

		$ip = $_SERVER['REMOTE_ADDR'];
		$r = rg_totp_enroll($db, $rg['login_ui']['uid'],
			$name, $secret, $ip, 't');
		if ($r !== TRUE) {
			$errmsg[] = rg_totp_error();
			break;
		}

		$ret .= rg_template('user/settings/totp/enroll_ok.html',
			$rg, TRUE /*xss*/);

		break;
	}

	$ret .= rg_totp_list_high_level($db, $rg);

	// defaults
	if ($enroll == 0) {
		$name = '';
		$ver = '';
		$secret = rg_totp_base32_generate(16);
	}

	$rg['totp']['name'] = $name;
	$rg['totp']['ver'] = $ver;
	$rg['totp']['secret'] = $secret;

	$png = rg_totp_png($rg['totp']['secret']);
	if ($png === FALSE) {
		$rg['totp']['img'] = 0;
	} else {
		$rg['totp']['img'] = 1;
		$rg['totp']['png'] = base64_encode($png);
	}

	$rg['HTML:errmsg'] = rg_template_errmsg($errmsg);

	$rg['rg_form_token'] = rg_token_get($db, $rg, 'user_totp');
	$ret .= rg_template('user/settings/totp/main.html', $rg, TRUE /*xss*/);

	// hints
	$hints = array();
	$hints[]['HTML:hint'] = rg_template("user/settings/totp/hints.html", $rg, FALSE /*xss*/);
	$ret .= rg_template_table("hints/list", $hints, $rg);

	rg_log_exit();
	rg_prof_end('totp_high_level');
	return $ret;
}

?>
